import datetime  # for date checking function
import re
import uuid
from enum import Enum
from pathlib import Path

import baseconv  # for the UUID
import packaging.version
import stix2.exceptions
from stix2 import FileSystemSource
from stix2 import Filter
import plyara.utils

from yara_validator.constants import MITRE_STIX_DATA_PATH
from yara_validator.stix2_patch.filter_casefold import FilterCasefold

METADATA = 'metadata'
BASE62_REGEX = r'^[0-9a-zA-z]+$'
UNIVERSAL_REGEX = r'^[^a-z]*$'
MITRE_GROUP_NAME = 'name'
CHILD_PLACE_HOLDER = 'child_place_holder'
DATE_FORMATS = [
    "%Y-%m", "%Y.%m", "%Y/%m",
    "%m/%d/%Y", "%d/%m/%Y",
    "%m/%d/%y", "%d/%m/%y",
    "%Y/%d/%m", "%y/%d/%m",
    "%Y/%m/%d", "%y/%m/%d",
    "%Y.%d.%m", "%y.%d.%m",
    "%Y.%m.%d", "%y.%m.%d",
    "%d-%m-%Y", "%m-%d-%Y", "%m-%d-%y", "%Y-%m-%d",
    "%d.%m.%Y", "%m.%d.%Y", "%m.%d.%y", "%Y.%m.%d",
    "%f/%e/%Y", "%f/%e/%y", "%e/%f/%Y", "%e/%f/%y",
    "%f-%e-%Y", "%f-%e-%y", "%e-%f-%Y", "%e-%f-%y",
    "%f.%e.%Y", "%f.%e.%y", "%e.%f.%Y", "%e.%f.%y",
    "%b %e, %Y", "%B %e, %Y",
    "%b %d, %Y", "%B %d, %Y",
    "%b %e %Y", "%B %e %Y", "%e %b %Y", "%e %B %Y",
    "%b %d %Y", "%B %d %Y", "%d %b %Y", "%d %B %Y",
    "%Y-%m-%d %I:%M:%S %p", "%Y/%m/%d %I:%M:%S %p"
    "%Y-%m-%d %H:%M:%S", "%Y/%m/%d %H:%M:%S"
]


# potential values of MetadataAttributes.optional variable
class MetadataOpt(Enum):
    REQ_PROVIDED = 'req_provided'
    REQ_OPTIONAL = 'req_optional'
    OPT_OPTIONAL = 'opt_optional'


# potential values for the encoding check
class StringEncoding(Enum):
    ASCII = 'ascii'
    UTF8 = 'utf-8'
    NONE = 'none'


def __is_ascii(rule_string):
    """
    Takes the string of the rule and parses it to check if there are only ascii characters present.
    :param rule_string: the string representation of the yara rule
    :return: true if there are only ascii characters in the string
    """
    return len(rule_string) == len(rule_string.encode())


def __is_utf8(rule_string):
    """
    Takes the string of the rule and parses it to check if there are only utf-8 characters present.
    :param rule_string: the string representation of the yara rule
    :return: true if there are only utf-8 characters in the string
    """
    try:
        rule_string.encode('utf-8')
    except UnicodeEncodeError:
        return False
    else:
        return True


def check_encoding(rule_string, encoding_flag):
    if encoding_flag == StringEncoding.ASCII.value:
        if not __is_ascii(rule_string):
            return False
    elif encoding_flag == StringEncoding.UTF8.value:
        if not __is_utf8(rule_string):
            return False

    return True


def convert_date(date):
    # Try a variety of known datetime formats
    # Return success in conversion as well as new formatted date

    for datetime_format in DATE_FORMATS:
        # Try a variety of known datetime formats
        try:
            parsed_date = datetime.datetime.strptime(date, datetime_format)
            return True, parsed_date.strftime("%Y-%m-%d")
        except ValueError:
            continue
    return False, date


class Validators:
    def __init__(self):
        self.required_fields = None
        self.required_fields_index = None
        self.required_fields_children = None
        self.category_types = None
        self.mitre_group_alias = None
        self.mitre_software_code = None
        self.names = {
            'None': self.valid_none,
            'valid_regex': self.valid_regex,
            'valid_uuid': self.valid_uuid,
            'valid_fingerprint': self.valid_fingerprint,
            'valid_version': self.valid_version,
            'valid_date': self.valid_date,
            'valid_source': self.valid_source,
            'valid_category': self.valid_category,
            'valid_category_type': self.valid_category_type,
            'valid_mitre_att': self.valid_mitre_att,
            'valid_actor': self.valid_actor,
            'mitre_group_generator': self.mitre_group_generator,
            'valid_percentage': self.valid_percentage
        }

    def update(self, required_fields, required_fields_index, required_fields_children, category_types):
        self.required_fields = required_fields
        self.required_fields_index = required_fields_index
        self.required_fields_children = required_fields_children
        self.category_types = category_types
        self.mitre_group_alias = None

    def valid_none(self, rule_to_validate, metadata_index, metadata_key, alias=None):
        self.required_fields[metadata_key].attributefound()
        self.required_fields_index[self.required_fields[metadata_key].position].increment_count()
        self.required_fields[metadata_key].attributevalid()
        return True

    def valid_percentage(self, rule_to_validate, metadata_index, metadata_key, alias=None):
        """
        Validates the metadata value is a valid percentage (within the range [0, 100])
        :param rule_to_validate: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the id metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value of the metadata value follows the regex expression or
            False if the value is does not match the expression
        """

        try:
            value = int(list(rule_to_validate[METADATA][metadata_index].values())[0])

            self.required_fields[metadata_key].attributefound()
            self.required_fields_index[self.required_fields[metadata_key].position].increment_count()

            if value <= 100 and value >= 0:
                self.required_fields[metadata_key].attributevalid()
            else:
                self.required_fields[metadata_key].attributeinvalid()
                return False
            return True
        except Exception:
            self.required_fields[metadata_key].attributeinvalid()
            return False

    def valid_regex(self, rule_to_validate, metadata_index, metadata_key, alias=None):
        """
        Validates the metadata value using provided regex expression
        :param rule_to_validate: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the id metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value of the metadata value follows the regex expression or
            False if the value is does not match the expression
        """
        value = list(rule_to_validate[METADATA][metadata_index].values())[0]

        if METADATA == 'hash':
            # Strip any whitespace before validation
            value = str(value).strip()
            rule_to_validate[METADATA].insert(metadata_index, {METADATA: value})

        self.required_fields[metadata_key].attributefound()
        self.required_fields_index[self.required_fields[metadata_key].position].increment_count()

        regex_expression = self.required_fields[metadata_key].argument.get('regexExpression')

        if re.fullmatch(regex_expression, value):
            self.required_fields[metadata_key].attributevalid()
        elif re.fullmatch(regex_expression, str(value).upper()):
            self.required_fields[metadata_key].attributevalid()
            rule_to_validate[METADATA][metadata_index][metadata_key] = str(value).upper()
        else:
            self.required_fields[metadata_key].attributeinvalid()
            return False
        return True

    def valid_uuid(self, rule_to_generate_uuid, metadata_index, metadata_key, alias=None):
        """
        Creates a valid UUID for the id metadata value and inserts it or verifies an existing id metadata value
        :param rule_to_generate_uuid: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the id metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if a the value of the id metadata value is of the correct size or if a new UUID is generated or
            False if the existing value is not of the correct size
        """
        UUID = metadata_key
        self.required_fields[UUID].attributefound()
        self.required_fields_index[self.required_fields[UUID].position].increment_count()

        rule_uuid = {UUID: str(baseconv.base62.encode(uuid.uuid4().int))}
        if Helper.valid_metadata_index(rule_to_generate_uuid, metadata_index):
            if list(rule_to_generate_uuid[METADATA][metadata_index].keys())[0] == UUID:
                if Helper.validate_uuid(list(rule_to_generate_uuid[METADATA][metadata_index].values())[0]):
                    self.required_fields[UUID].attributevalid()
                else:
                    self.required_fields[UUID].attributeinvalid()
            else:
                rule_to_generate_uuid[METADATA].insert(metadata_index, rule_uuid)
                self.required_fields[UUID].attributevalid()
        else:
            rule_to_generate_uuid[METADATA].append(rule_uuid)
            self.required_fields[UUID].attributevalid()

        return self.required_fields[UUID].valid

    def valid_fingerprint(self, rule_to_generate_id, metadata_index, metadata_key, alias=None):
        """
        Calculates a valid fingerprint for the fingerprint metadata value and inserts it or replaces the existing value
            of the fingerprint metadata value.
            Current functionality is not to check the value of an existing fingerprint metadata value and just overwrite
            it as this is automatically filled out.
        :param rule_to_generate_id: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the fingerprint metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: This should return True all the time as there will always be a return from self.calculate_rule_hash
        """
        FINGERPRINT = metadata_key
        self.required_fields[FINGERPRINT].attributefound()
        self.required_fields_index[self.required_fields[FINGERPRINT].position].increment_count()

        rule_hash = plyara.utils.generate_hash(rule_to_generate_id)
        if rule_hash:
            rule_id = {FINGERPRINT: rule_hash}
            if Helper.valid_metadata_index(rule_to_generate_id, metadata_index):
                if list(rule_to_generate_id[METADATA][metadata_index].keys())[0] == FINGERPRINT:
                    rule_to_generate_id[METADATA][metadata_index] = rule_id
                    self.required_fields[FINGERPRINT].attributevalid()
                else:
                    rule_to_generate_id[METADATA].insert(metadata_index, rule_id)
                    self.required_fields[FINGERPRINT].attributevalid()
            else:
                rule_to_generate_id[METADATA].append(rule_id)
                self.required_fields[FINGERPRINT].attributevalid()

        return self.required_fields[FINGERPRINT].valid

    def valid_version(self, rule_to_version_check, metadata_index, metadata_key, alias=None):
        """
        This value can be generated: there is the option to verify if an existing version format is correct, insert a
            generated version if none was found and if the potential default metadata index would be out of bounds
            appends a generated version
        :param rule_to_version_check: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the version metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the version is of the correct format and False if it is not
        """
        VERSION = metadata_key
        self.required_fields[VERSION].attributefound()
        self.required_fields_index[self.required_fields[VERSION].position].increment_count()

        rule_version = {VERSION: '1.0'}
        if Helper.valid_metadata_index(rule_to_version_check, metadata_index):
            if list(rule_to_version_check[METADATA][metadata_index].keys())[0] == VERSION:
                if isinstance(packaging.version.parse(
                        list(rule_to_version_check[METADATA][metadata_index].values())[0]),
                        packaging.version.Version):
                    self.required_fields[VERSION].attributevalid()
                else:
                    self.required_fields[VERSION].attributeinvalid()
            else:
                rule_to_version_check[METADATA].insert(metadata_index, rule_version)
                self.required_fields[VERSION].attributevalid()
        else:
            rule_to_version_check[METADATA].append(rule_version)
            self.required_fields[VERSION].attributevalid()

        return self.required_fields[VERSION].valid

    def valid_date(self, rule_to_date_check, metadata_index, metadata_key, alias=None):
        """
        This value can be generated: there is the option to verify if an existing date is correct, insert a generated
            date if none was found and if the potential default metadata index would be out of bounds appends
            a generated date
        :param rule_to_date_check: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the last_modified metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value matches the valid date format and False if it does not match it
        """
        def update_metadata(data, force=False):
            if alias or force:
                # Overwrite with proper metadata field
                rule_date = {DATE: data}
                rule_to_date_check[METADATA].pop(metadata_index)
                rule_to_date_check[METADATA].insert(metadata_index, rule_date)

        DATE = metadata_key
        if alias:
            DATE = alias
        self.required_fields[DATE].attributefound()
        self.required_fields_index[self.required_fields[DATE].position].increment_count()

        if Helper.valid_metadata_index(rule_to_date_check, metadata_index):
            if list(rule_to_date_check[METADATA][metadata_index].keys())[0] == metadata_key:
                date = list(rule_to_date_check[METADATA][metadata_index].values())[0]
                if Helper.validate_date(date):
                    self.required_fields[DATE].attributevalid()
                    update_metadata(date)
                else:
                    # Date isn't in the expected format, let's see if we can convert it
                    success, formatted_date = convert_date(date)
                    if success and Helper.validate_date(formatted_date):
                        self.required_fields[DATE].attibutevalid()
                        update_metadata(date, force=True)
                    else:
                        self.required_fields[DATE].attributeinvalid()
            else:
                rule_date = {DATE: Helper.current_valid_date()}
                rule_to_date_check[METADATA].insert(metadata_index, rule_date)
                self.required_fields[DATE].attributevalid()
        else:
            rule_date = {DATE: Helper.current_valid_date()}
            rule_to_date_check[METADATA].append(rule_date)
            self.required_fields[DATE].attributevalid()

        return self.required_fields[DATE].valid

    def valid_source(self, rule_to_source_check, metadata_index, metadata_key, alias=None):
        """
        Validates the source
        :param rule_to_source_check:
        :param metadata_index: used to reference what the array index of the source metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value matches the UNIVERSAL_REGEX and False if it does not match it
        """
        SOURCE = metadata_key
        self.required_fields[SOURCE].attributefound()
        self.required_fields_index[self.required_fields[SOURCE].position].increment_count()

        metadata = rule_to_source_check[METADATA]
        source_to_check = metadata[metadata_index][SOURCE]
        if re.fullmatch(UNIVERSAL_REGEX, source_to_check):
            self.required_fields[SOURCE].attributevalid()
        elif re.fullmatch(UNIVERSAL_REGEX, str(source_to_check).upper()):
            source_to_check = str(source_to_check).upper()
            metadata[metadata_index][SOURCE] = source_to_check
            self.required_fields[SOURCE].attributevalid()
        else:
            self.required_fields[SOURCE].attributeinvalid()

        return self.required_fields[SOURCE].valid

    def valid_mitre_att(self, rule_to_validate_mitre_att, metadata_index, metadata_key, alias=None):
        """
        Pulls the value of the mitre_att metadata value and passes it to validate_mitre_att_by_id. This also stores
            any found MITRE ATT&CK software codes to be compared with ones generated by malware metadata values
        :param rule_to_validate_mitre_att: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the mitre_att metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value was found in the MITRE ATT&CK database and False if it was not found
        """
        MITRE_ATT = metadata_key
        MITRE_SOFTWAREID_FOUND = 'mitre_softwareid_found'
        self.required_fields[MITRE_ATT].attributefound()
        self.required_fields_index[self.required_fields[MITRE_ATT].position].increment_count()

        metadata = rule_to_validate_mitre_att[METADATA]
        mitre_att_to_validate = str(metadata[metadata_index][MITRE_ATT]).upper()
        metadata[metadata_index][MITRE_ATT] = mitre_att_to_validate
        if Helper.validate_mitre_att_by_id(mitre_att_to_validate):
            self.required_fields[MITRE_ATT].attributevalid()
        else:
            self.required_fields[MITRE_ATT].attributeinvalid()

        if self.required_fields[MITRE_ATT].valid and mitre_att_to_validate.startswith('S'):
            soft_codes_found = self.required_fields[MITRE_ATT].check_argument_list_var(MITRE_SOFTWAREID_FOUND)
            soft_codes_found.append(mitre_att_to_validate)

        return self.required_fields[MITRE_ATT].valid

    def valid_category(self, rule_to_validate_category, metadata_index, metadata_key, alias=None):
        """
        Pulls the value of the category metadata value and checks if it is a valid category type.
            Valid options are stored in self.category_types. If the category value is valid and a new metadata
            metadata with a name the same as the category value is added to be searched for.
            This new metadata value links to the same object as the initially created
            self.required_fields[CATEGORY_TYPE].
        :param rule_to_validate_category: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the category metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value was found in self.category_types and False if it was not found
        """
        CATEGORY = metadata_key
        self.required_fields[CATEGORY].attributefound()
        self.required_fields_index[self.required_fields[CATEGORY].position].increment_count()
        child_metadata_place_holder = self.required_fields[CATEGORY].argument.get(CHILD_PLACE_HOLDER)

        metadata = rule_to_validate_category[METADATA]
        rule_category_to_check = metadata[metadata_index][CATEGORY]
        if rule_category_to_check in self.category_types:
            self.required_fields[CATEGORY].attributevalid()
            add_category_type_to_required = {
                str(rule_category_to_check).lower(): self.required_fields[child_metadata_place_holder]}
            self.required_fields_children.update(add_category_type_to_required)
        elif str(rule_category_to_check).upper() in self.category_types:
            rule_category_to_check = str(rule_category_to_check).upper()
            metadata[metadata_index][CATEGORY] = rule_category_to_check
            self.required_fields[CATEGORY].attributevalid()
            add_category_type_to_required = {
                str(rule_category_to_check).lower(): self.required_fields[child_metadata_place_holder]}
            self.required_fields_children.update(add_category_type_to_required)
        else:
            self.required_fields[CATEGORY].attributeinvalid()

        return self.required_fields[CATEGORY].valid

    def valid_category_type(self, rule_to_validate_type, metadata_index, metadata_key, alias=None):
        """
        This will be called by the new metadata created by the valid_category function. Because it references the same
            object as that initialized as CATEGORY_TYPE we can use that to reference the reqired metadata in this
            function. This also generates and stores any MITRE ATT&CK software codes to be checked against found
            mitre_att software codes
        :param rule_to_validate_type: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the category_type metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value matches the Regex expression and False if it was not found
        """
        CATEGORY = 'category'
        GENERATE_MITRE_ATT_FROM = 'generate_mitre_att_from'
        MITRE_SOFTWAREID_GEN = 'mitre_softwareid_gen'
        child_metadata_place_holder = self.required_fields[CATEGORY].argument.get(CHILD_PLACE_HOLDER)
        self.required_fields[child_metadata_place_holder].attributefound()
        self.required_fields_index[self.required_fields[child_metadata_place_holder].position].increment_count()
        key_gen_mitre_att = r'' +\
                            self.required_fields[child_metadata_place_holder].argument.get(GENERATE_MITRE_ATT_FROM)

        metadata = rule_to_validate_type[METADATA]
        rule_category_key_to_check = list(metadata[metadata_index].keys())[0]
        rule_category_value_to_check = list(metadata[metadata_index].values())[0]
        if re.fullmatch(UNIVERSAL_REGEX, rule_category_value_to_check):
            self.required_fields[child_metadata_place_holder].attributevalid()
        elif re.fullmatch(UNIVERSAL_REGEX, str(rule_category_value_to_check).upper()):
            rule_category_value_to_check = str(rule_category_value_to_check).upper()
            metadata[metadata_index][rule_category_key_to_check] = rule_category_value_to_check
            self.required_fields[child_metadata_place_holder].attributevalid()
        else:
            self.required_fields[child_metadata_place_holder].attributeinvalid()

        rule_category_value_to_check = list(metadata[metadata_index].values())[0]
        if self.required_fields[child_metadata_place_holder].valid and re.fullmatch(key_gen_mitre_att,
                                                                                    rule_category_key_to_check):
            malware_id = Helper.get_software_id_by_name(rule_category_value_to_check)
            if malware_id:
                malware_ids_found = self.required_fields[child_metadata_place_holder] \
                                        .check_argument_list_var(MITRE_SOFTWAREID_GEN)
                malware_ids_found.append(malware_id)

        return self.required_fields[child_metadata_place_holder].valid

    def valid_actor(self, rule_to_validate_actor, metadata_index, metadata_key, alias=None):
        """
        Validates the actor, makes the actor_type metadata value required.
            Adds a required metadata value for mitre_group to hold the a potential alias value.
            Also stores the value of the actor metadata value in self.mitre_group_alias variable for use with the
            mitre_group_generator function
        :param rule_to_validate_actor: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the actor metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: True if the value matches the self.mitre_group_alias_regex and False if it does not
        """
        ACTOR = metadata_key
        ACTOR_TYPE = self.required_fields[ACTOR].argument.get('required')
        child_metadata = self.required_fields[ACTOR].argument.get('child')
        child_metadata_place_holder = self.required_fields[ACTOR].argument.get(CHILD_PLACE_HOLDER)
        mitre_group_alias_regex = r'^[^a-z]+$'

        self.required_fields[ACTOR].attributefound()
        self.required_fields_index[self.required_fields[ACTOR].position].increment_count()

        metadata = rule_to_validate_actor[METADATA]
        actor_to_check = metadata[metadata_index][ACTOR]
        if re.fullmatch(mitre_group_alias_regex, actor_to_check):
            self.required_fields[ACTOR].attributevalid()
            add_mitre_group_to_required = {child_metadata: self.required_fields[child_metadata_place_holder]}
            self.required_fields_children.update(add_mitre_group_to_required)
            self.mitre_group_alias = actor_to_check
        elif re.fullmatch(mitre_group_alias_regex, str(actor_to_check).upper()):
            actor_to_check = str(actor_to_check).upper()
            metadata[metadata_index][ACTOR] = actor_to_check
            self.required_fields[ACTOR].attributevalid()
            add_mitre_group_to_required = {child_metadata: self.required_fields[child_metadata_place_holder]}
            self.required_fields_children.update(add_mitre_group_to_required)
            self.mitre_group_alias = actor_to_check
        else:
            self.required_fields[ACTOR].attributeinvalid()

        return self.required_fields[ACTOR].valid

    def mitre_group_generator(self, rule_to_generate_group, metadata_index, metadata_key, alias=None):
        """
        This will only be looked for if the actor metadata value has already been processed.
            Current functionality is not to check the value of an existing mitre_group metadata value and just overwrite
            it as this is automatically filled out. Also if no alias is found it will be removed.
        :param rule_to_generate_group: the plyara parsed rule that is being validated
        :param metadata_index: used to reference what the array index of the mitre_group metadata value is
        :param metadata_key: the name of the metadata value that is being processed
        :return: This should return True all the time as there will always be a return from self.get_group_from_alias
        """
        ACTOR = 'actor'
        place_holder = self.required_fields[ACTOR].argument.get(CHILD_PLACE_HOLDER)
        if self.required_fields.get(metadata_key):  # if child place holder is passed as metadata_key
            MITRE_GROUP = self.required_fields[self.required_fields[metadata_key].argument['parent']].argument['child']
        else:
            MITRE_GROUP = metadata_key

        mitre_group = str(Helper.get_group_from_alias(self.mitre_group_alias)).upper()
        rule_group = {MITRE_GROUP: mitre_group}
        if Helper.valid_metadata_index(rule_to_generate_group, metadata_index):
            if list(rule_to_generate_group[METADATA][metadata_index].keys())[0] == MITRE_GROUP:
                if mitre_group:
                    rule_to_generate_group[METADATA][metadata_index] = rule_group
                    self.required_fields[place_holder].attributefound()
                    self.required_fields[place_holder].attributevalid()
                    self.required_fields_index[self.required_fields[place_holder].position].increment_count()
                else:
                    rule_to_generate_group[METADATA].pop(metadata_index)
                    return True
            else:
                if mitre_group:
                    rule_to_generate_group[METADATA].insert(metadata_index, rule_group)
                    self.required_fields[place_holder].attributefound()
                    self.required_fields[place_holder].attributevalid()
                    self.required_fields_index[self.required_fields[place_holder].position].increment_count()
                else:
                    return True
        else:
            if mitre_group:
                rule_to_generate_group[METADATA].append(rule_group)
                self.required_fields[place_holder].attributefound()
                self.required_fields[place_holder].attributevalid()
                self.required_fields_index[self.required_fields[place_holder].position].increment_count()
            else:
                return True

        return self.required_fields[place_holder].valid

    def mitre_software_generator(self, rule_to_generate_mitre_att, category_key, mitre_key, alias=None):
        """
        This is called to generate any required mitre_att software codes. This is determined by comparing software
            codes generated from the value of malware metadata keys and any already found mitre_att keys containing
            software codes. Any of the generated values that are not found are added to the metadata of the rule
            as mitre_att
        :param rule_to_generate_mitre_att: rule that we want to add any missing mitre_att software codes
        :param category_key: the key for category, which is used to find the generated software codes
        :param mitre_key: the key used to find the found software codes
        :return:
        """
        CATEGORY = category_key
        MITRE_ATT = mitre_key
        MITRE_SOFTWAREID_GEN = 'mitre_softwareid_gen'
        MITRE_SOFTWAREID_FOUND = 'mitre_softwareid_found'
        child_metadata_place_holder = self.required_fields[CATEGORY].argument.get(CHILD_PLACE_HOLDER)

        malware_ids_found = self.required_fields[child_metadata_place_holder]\
                                .check_argument_list_var(MITRE_SOFTWAREID_GEN)
        soft_codes_found = self.required_fields[MITRE_ATT].check_argument_list_var(MITRE_SOFTWAREID_FOUND)

        for malware_id_found in malware_ids_found:
            if malware_id_found not in soft_codes_found:
                mitre_att_to_add = {MITRE_ATT: malware_id_found}
                rule_to_generate_mitre_att[METADATA].append(mitre_att_to_add)
                self.required_fields[MITRE_ATT].attributefound()
                self.required_fields[MITRE_ATT].attributevalid()
                self.required_fields_index[self.required_fields[MITRE_ATT].position].increment_count()


class Helper:
    import os
    if not os.path.exists(MITRE_STIX_DATA_PATH):
        from git import Repo
        print(f'Unable to find STIX data on {MITRE_STIX_DATA_PATH}. Cloning..')

        os.makedirs(MITRE_STIX_DATA_PATH)
        repo = Repo.clone_from('https://github.com/mitre/cti.git', to_path=MITRE_STIX_DATA_PATH, depth=1,
                               branch="ATT&CK-v13.1")
    fs = FileSystemSource(os.path.join(MITRE_STIX_DATA_PATH, 'enterprise-attack'))

    @staticmethod
    def valid_metadata_index(rule, index):
        """
        Ensures that the index will not return an out of bounds error
        :param rule: the plyara parsed rule that is being validated
        :param index: the potential index
        :return: True if the potential index will not be out of bounds and false otherwise
        """
        count_of_metadata = len(rule[METADATA])
        if index >= count_of_metadata:
            return False
        else:
            return True

    @staticmethod
    def validate_uuid(uuid_to_check):
        """
        Validates the uuid by checking the base62_uuid matches the potential characters and is of the correct length
        :param uuid_to_check: the value to be
        :return: True if the decoded value of the id is 127 bits in length and False otherwise
        """
        if re.fullmatch(BASE62_REGEX, uuid_to_check):
            return 20 <= len(uuid_to_check) <= 22
        else:
            return False

    @staticmethod
    def regex_match_string_names_for_values(string_name_preface, string_name_expression, string_substitutions):
        """
        Looks to replace YARA string references in conditions such as $a*. The function looks to match all matching
            string names and compile a completed list of those string values
        :param string_name_preface: Can be one of '$', '!', '#', or '@'
        :param string_name_expression: The string name expression that will be converted into a regex pattern
        :param string_substitutions: the dict of all string substitutions and values
        :return: the completed list of string values whose string names match the expression
        """
        string_name, string_suffix = string_name_expression[:-1], string_name_expression[-1:]
        string_name_regex = r'^\{}.{}$'.format(string_name, string_suffix)
        string_value_matches = []
        for key in string_substitutions.keys():
            if re.fullmatch(string_name_regex, key):
                string_value_matches.append(string_name_preface + string_substitutions[key])

        return string_value_matches

    @staticmethod
    def sort_strings_add_commas(list_of_strings):
        """
        Takes a list of string values and rebuilds it as a string with comma delimiters so it would look like a hard
            coded YARA list of strings
        :param list_of_strings: the list of collected string values
        :return: the sorted list
        """
        list_of_strings.sort()
        count_of_strings = len(list_of_strings)

        for index, string in enumerate(list_of_strings):
            if index + 1 < count_of_strings:
                list_of_strings.insert(index + index + 1, ',')

        return list_of_strings

    @staticmethod
    def validate_date(date_to_validate):
        """
        Verifies a date is in the correct format.
        :param date_to_validate: the value of the last_modifed metadata value
        :return: True if the value is in the correct format or False if it is not valid
        """
        try:
            if date_to_validate != datetime.datetime.strptime(date_to_validate, '%Y-%m-%d').strftime('%Y-%m-%d'):
                raise ValueError
            return True
        except ValueError:
            return False

    @staticmethod
    def current_valid_date():
        """
        Generates the current date in the valid format
        :return: the current date in the valid format
        """
        return datetime.datetime.now().strftime('%Y-%m-%d')

    @staticmethod
    def get_tactic_by_id(id_code):
        """
        Used if the id_code is prefaced with TA
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        return Helper.fs.query([
            Filter('type', '=', 'x-mitre-tactic'),
            Filter('external_references.external_id', '=', id_code)
        ])

    @staticmethod
    def get_technique_by_id(id_code):
        """
        Used if the id_code is prefaced with T
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        return Helper.fs.query([
            Filter('type', '=', 'attack-pattern'),
            Filter('external_references.external_id', '=', id_code)
        ])

    @staticmethod
    def get_software_by_id(id_code):
        """
        Used if the id_code is prefaced with S
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        malware_return = Helper.fs.query([
            Filter('type', '=', 'malware'),
            Filter('external_references.external_id', '=', id_code)
        ])

        tool_return = Helper.fs.query([
            Filter('type', '=', 'tool'),
            Filter('external_references.external_id', '=', id_code)
        ])

        if malware_return:
            return malware_return
        elif tool_return:
            return tool_return

    @staticmethod
    def get_group_by_id(id_code):
        """
        Used if the id_code is prefaced with G
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        return Helper.fs.query([
            Filter('type', '=', 'intrusion-set'),
            Filter('external_references.external_id', '=', id_code)
        ])

    @staticmethod
    def get_mitigation_by_id(id_code):
        """
        Used if the id_code is prefaced with M
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        return Helper.fs.query([
            Filter('type', '=', 'course-of-action'),
            Filter('external_references.external_id', '=', id_code)
        ])

    @staticmethod
    def get_mitreattck_by_id(id_code):
        """
        Used if the id_code is prefaced with an unknown letter. This is about 20x more inefficient then the queries
            that specify types in the filters
            It is used as a catch all in case new MITRE ATT&CK ID Code types are added in the future
        :param id_code: The value of the mitre_att metadata value
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        try:
            Helper.fs.query([
                Filter('external_references.external_id', '=', id_code)
            ])
        except stix2.exceptions.InvalidValueError:
            return ''

    @staticmethod
    def validate_mitre_att_by_id(id_code):
        """
        Checks the preface of the id_code and sends the id_code to specific functions
            This is done because using specified filters based on the type is about 20 times more efficient then
            the entire MITRE ATT&CK database with the id_code.
            There is a catch all provided in the case that there are new ID Code types added in the future
        :param id_code: The value of the mitre_att metadata value
        :return: The return from the specified get_ function
        """
        if id_code.startswith('TA'):
            return Helper.get_tactic_by_id(id_code)
        elif id_code.startswith('T'):
            return Helper.get_technique_by_id(id_code)
        elif id_code.startswith('S'):
            return Helper.get_software_by_id(id_code)
        elif id_code.startswith('G'):
            return Helper.get_group_by_id(id_code)
        elif id_code.startswith('M'):
            return Helper.get_mitigation_by_id(id_code)
        else:
            return Helper.get_mitreattck_by_id(id_code)

    @staticmethod
    def get_group_from_alias(alias):
        """
        Maps any alias to the potential MITRE ATT&CK group name, if the provided name is a known alias.
        :param alias: The alias to check
        :return: Either returns the MITRE ATT&CK group name or returns empty string, '', if the query returns null
        """
        group_from_alias = Helper.fs.query([
            Filter('type', '=', 'intrusion-set'),
            FilterCasefold('aliases', 'casefold', alias)
        ])

        if not group_from_alias:
            return ''

        return group_from_alias[0][MITRE_GROUP_NAME]

    @staticmethod
    def get_software_id_by_name(software_name):
        """
        Used to find the software id with the software_name
        :param software_name: The value of the malware name
        :return: The return of the query of the MITRE ATT&CK database, null if there are no matches
        """
        malware_return = Helper.fs.query([
            Filter('type', '=', 'malware'),
            FilterCasefold('name', 'casefold', software_name)
        ])

        tool_return = Helper.fs.query([
            Filter('type', '=', 'tool'),
            FilterCasefold('name', 'casefold', software_name)
        ])

        if malware_return:
            return malware_return[0]['external_references'][0]['external_id']
        elif tool_return:
            return tool_return[0]['external_references'][0]['external_id']
        else:
            return ''



import plyara.utils
import collections
import re
import yara
from pathlib import Path

SYNTAX_ERRORS_INCLUDES = re.compile(r'.*can\'t open include file')
SYNTAX_ERRORS_UNDEFINED_MODULES = re.compile(
    r'.*undefined identifier (?="pe"|"elf"|"cuckoo"|"magic"|"hash"|"math"|"dotnet"|"time")')
SYNTAX_ERRORS_UNDEFINED = re.compile(r'.*undefined identifier')

class YaraFileProcessor:
    """
    YaraFileProcessor class is used to process a given rule file and parse it into one or more YARA rules
    """

    def __init__(self, rule_file, char_to_replace, char_replacement, count_of_replaced,
                 check_import_modules=True, check_rule_dependancies=False):
        # Original rule file
        self.original_rule_file = rule_file
        # Variables for the white space standardization
        self.char_to_replace = char_to_replace.encode('utf-8').decode('unicode_escape')
        self.char_replacement = char_replacement.encode('utf-8').decode('unicode_escape')
        self.count_of_replaced = count_of_replaced
        # String representation to contain edits to the original rule
        self.edited_rule_string = ''
        # Array to contain the YARA rules
        self.yara_rules = []
        # Overall rule error flag
        self.file_errors = False
        # Overall warning flag
        self.file_warnings = False
        # collection of all the file errors
        self.errors = collections.OrderedDict()
        # collection of all the file warnings (not used yet)
        self.warnings = collections.OrderedDict()

        # Plyara object for parsing the yara rule file
        parser = plyara.Plyara()
        file_name = ''
        # This block attempts to read the file as utf-8. If there are any issues with the file format or reading
        #   the file it creates a yara_rule object and sets the error state
        if isinstance(rule_file, str) or isinstance(rule_file, Path):
            if isinstance(rule_file, str):
                file_as_path = Path(self.original_rule_file)
            else:
                file_as_path = self.original_rule_file
            file_name = file_as_path.name
            with open(rule_file, encoding='utf-8') as yf:
                try:
                    self.original_rule_string = yf.read()
                except UnicodeDecodeError as e:
                    print('UnicodeDecodeError: ' + str(e))
                    file_response = 'UnicodeDecodeError:\t{!r}'.format(str(e))
                    self.update_file_error(True, str(file_name), file_response)
                    return
                except Exception as e:
                    print('There was an error opening the file: ' + str(e))
                    file_response = 'There was an error opening the file:\t{!r}'.format(str(e))
                    self.update_file_error(True, str(file_name), file_response)
                    return
        else:
            self.original_rule_string = rule_file
            file_name = 'File_Contents'

        # This block attempts to compile the self.original_rule_string. If there are any issues compiling the file it
        #   creates a yara_rule objects and sets the error state
        try:
            rules = yara.compile(source=self.original_rule_string)
        except yara.SyntaxError as e:
            file_response = None
            error_string = str(e)
            if SYNTAX_ERRORS_INCLUDES.match(error_string):
                pass
            elif SYNTAX_ERRORS_UNDEFINED_MODULES.match(error_string):
                if check_import_modules:
                    # print('Error Compiling YARA file with yara: ' + str(e))
                    file_response = 'Error Compiling YARA file with yara:\t{!r}'.format(str(e))
            elif SYNTAX_ERRORS_UNDEFINED.match(error_string):
                if check_rule_dependancies:
                    # print('Error Compiling YARA file with yara: ' + str(e))
                    file_response = 'Error Compiling YARA file with yara:\t{!r}'.format(str(e))
            else:
                # print('Error Compiling YARA file with yara: ' + str(e))
                file_response = 'Error Compiling YARA file with yara:\t{!r}'.format(str(e))

            if file_response:
                self.update_file_error(True, str(file_name), file_response)
                return

        except Exception as e:
            # print('Error Compiling YARA file with yara: ' + str(e))
            file_response = 'Error Compiling YARA file with yara:\t{!r}'.format(str(e))
            self.update_file_error(True, str(file_name), file_response)
            return

        # This block attempts to parse the self.original_rule_string. If there are any issues parsing the file it a
        #   yara_rule object and sets the error state
        try:
            self.plyara_rule = parser.parse_string(self.original_rule_string)
        except plyara.exceptions.ParseTypeError as e:
            # print('Error reported by plyara library: plyara.exceptions.ParseTypeError: ' + str(e))
            file_response = 'Error reported by plyara library: plyara.exceptions.ParseTypeError:\t{!r}'.format(str(e))
            self.update_file_error(True, str(file_name), file_response)
            return
        except Exception as e:
            # print('Error Parsing YARA file with plyara: ' + str(e))
            file_response = 'Error Parsing YARA file with plyara:\t{!r}'.format(str(e))
            self.update_file_error(True, str(file_name), file_response)
            return

        # The number of rules found in the file
        self.count_of_rules = len(self.plyara_rule)
        # Process the string and PlYara into an array of YaraRule objects
        self.__process_rule_representiations_to_array()

    def __process_rule_representiations_to_array(self):
        """
        Processes the contents of the rule file into an array of YaraObjects, one object per yara rule found
            Each YaraObject contains the string representation of the rule, the plyara representation and
            a rule reutrn object
        :return:
        """
        if self.count_of_rules > 0:
            yara_rule_split_line = self.original_rule_string.splitlines()
            for plyara_rule in self.plyara_rule:
                string_of_rule = yara_rule_split_line[plyara_rule['start_line'] - 1:plyara_rule['stop_line']]
                string_of_rule = "\n".join(string_of_rule)
                yara_rule = YaraRule(string_of_rule, plyara_rule)
                self.yara_rules.append(yara_rule)

    def __replace_for_each_one_to_many(self, line):
        """
        Takes a line, transforms it into a list, parses through the list looking for the self.char_to_replace character
            and replaces each instance found with self.char_replacement * self.count_of_replaced
        :param line: a line that starts with at least one self.char_to_replace_character
        :return:
        """
        new_list = []
        character_replace = [self.char_replacement] * self.count_of_replaced
        line_as_list = list(line)
        non_white_space_index = 0
        for index, character in enumerate(line_as_list):
            if re.match(self.char_to_replace, character):
                new_list = new_list + character_replace
            elif re.match(self.char_replacement, character):
                new_list.append(character)
            else:
                non_white_space_index = index
                break

        if non_white_space_index == 0:
            non_white_space_index = len(line_as_list)
        new_list = new_list + line_as_list[non_white_space_index:]

        newline = ''.join(new_list)
        return newline

    def __standardize_white_space(self, edited_rule_string):
        """
        Takes the edited_rule_string, scans the start of each line for the self.char_to_replace and passes any line
            found to start with that character to __replace_for_each_one_to_many
        :param edited_rule_string: the array of lines
        :return:
        """
        regex_of_char_to_replace = '^' + '[' + self.char_to_replace + self.char_replacement + ']' + '+'
        for index, line in enumerate(edited_rule_string):
            if re.match(regex_of_char_to_replace, line):
                edited_rule_string[index] = self.__replace_for_each_one_to_many(line)

    def strings_of_rules_to_original_file(self):
        """
        This rebuilds a rule string incorporating any changes from the rule return objects
        :return:
        """
        edited_rule_string = self.original_rule_string.splitlines()
        yara_rules_reversed = self.yara_rules
        yara_rules_reversed.reverse()
        for rule in yara_rules_reversed:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    changed_rule_string = rule.rule_return.edited_rule.splitlines()
                else:
                    changed_rule_string = rule.rule_return.validated_rule.splitlines()
                edited_rule_string = edited_rule_string[0:rule.rule_plyara['start_line'] - 1]\
                                     + changed_rule_string + edited_rule_string[rule.rule_plyara['stop_line']:]

        self.__standardize_white_space(edited_rule_string)
        edited_rule_string = '\n'.join(edited_rule_string)
        self.edited_rule_string = edited_rule_string

    def strings_of_rules_to_multi_files(self):
        """
        This will potentially rebuild a multi-rule yara file into an array of strings. Each string will represent a
            single yara rule which will incorporate any changes from the rule return objects.

            NO CONTENT YET
        :return:
        """

    def update_file_error(self, file_error, error_tag, message):
        if not self.file_errors:
            self.file_errors = file_error

        self.errors[error_tag] = message

    def update_file_warning(self, file_warning, warning_tag, message):
        if not self.file_warnings:
            self.file_warnings = file_warning

        self.warnings[warning_tag] = message

    def return_edited_file_string(self):
        return self.edited_rule_string

    def __build_return_string(self, collection):
        return_string = ''
        for index, metadata in enumerate(collection):
            if index > 0:
                return_string += '\n'
            return_string = '{}{}: {}'.format(return_string, metadata, collection[metadata])

        return return_string

    def __build_return_string_cmlt(self, collection):
        format_string = '{indent:>8} {key:30} {value}'
        return_string = '\n'.join([
            format_string.format(indent='-', key=k + ':', value=v)
            for k, v in collection.items()
        ])

        return return_string

    def return_file_error_state(self):
        """
        Loops through the self.yara_rules array and returns true if any of the rules are in an error state
        :return:
        """
        error_state = False
        if self.file_errors:
            error_state = self.file_errors
            return error_state

        for rule in self.yara_rules:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    if rule.return_error():
                        error_state = rule.return_error()
                        break
                else:
                    if not rule.rule_return.rule_validity:
                        error_state = not rule.rule_return.rule_validity
                        break

        return error_state

    def return_rule_errors(self):
        """
        Returns the any file errors and loops through the self.yara_rules array and returns a string for any errors
        :return: error_string, a string of all of the errors concatenated together
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_rule_errors() is deprecated, use YaraFileProcessor.return_file_errors() instead',
            DeprecationWarning
        )
        error_string = ''

        if self.file_errors:
            error_string = self.__build_return_string(self.errors)

        for rule in self.yara_rules:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    if rule.return_error():
                        error_string = error_string + rule.return_errors()
                else:
                    if not rule.rule_return.rule_validity:
                        error_string = error_string + rule.return_errors()

        return error_string

    def return_file_errors(self):
        """
        Returns the any file errors and loops through the self.yara_rules array and returns a string for any errors
        :return: error_string, a string of all of the errors concatenated together
        """
        error_string = ''

        if self.file_errors:
            error_string = self.__build_return_string(self.errors)

        for rule in self.yara_rules:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    if rule.return_error():
                        error_string = error_string + rule.return_errors()
                else:
                    if not rule.rule_return.rule_validity:
                        error_string = error_string + rule.return_errors()

        return error_string

    def return_rule_errors_for_cmlt(self):
        """
        Loops throught the self.yara_rules array and returns a string for of errors in cmlt format
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_rule_errors_for_cmlt() is deprecated, '
            'use YaraFileProcessor.return_file_errors_for_cmlt() instead',
            DeprecationWarning
        )
        error_string = ''

        if self.file_errors:
            error_string = self.__build_return_string_cmlt(self.errors)

        for rule in self.yara_rules:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    if rule.return_error():
                        error_string = error_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                        error_string = error_string + rule.rule_plyara["rule_name"] + "\n"
                        error_string = error_string + rule.return_errors_for_cmlt()
                else:
                    if not rule.rule_return.rule_validity:
                        error_string = error_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                        error_string = error_string + rule.return_errors_for_cmlt()

        return error_string

    def return_file_errors_for_cmlt(self):
        """
        Loops throught the self.yara_rules array and returns a string for of errors in cmlt format
        :return:
        """
        error_string = ''

        if self.file_errors:
            error_string = self.__build_return_string_cmlt(self.errors)

        for rule in self.yara_rules:
            if rule.rule_return:
                if isinstance(rule.rule_return, YaraReturn):
                    if rule.return_error():
                        error_string = error_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                        error_string = error_string + rule.rule_plyara["rule_name"] + "\n"
                        error_string = error_string + rule.return_errors_for_cmlt()
                else:
                    if not rule.rule_return.rule_validity:
                        error_string = error_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                        error_string = error_string + rule.return_errors_for_cmlt()

        return error_string

    def return_rule_warning_state(self):
        """
        Loops through the self.yara_rules array and returns true if any of the rules are in a warning state
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_rule_warning_state() is deprecated, '
            'use YaraFileProcessor.return_file_warning_state() instead',
            DeprecationWarning
        )
        warning_state = False
        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_state = True
                    break

        return warning_state

    def return_file_warning_state(self):
        """
        Loops through the self.yara_rules array and returns true if any of the rules are in a warning state
        :return:
        """
        warning_state = False
        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_state = True
                    break

        return warning_state

    def return_rule_warnings(self):
        """
        Loops throught the self.yara_rules array and returns a string for of warnings
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_rule_warnings() is deprecated, '
            'use YaraFileProcessor.return_file_warnings() instead',
            DeprecationWarning
        )
        warning_string = ''

        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_string = warning_string + rule.return_warnings()

        return warning_string

    def return_file_warnings(self):
        """
        Loops throught the self.yara_rules array and returns a string for of warnings
        :return:
        """
        warning_string = ''

        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_string = warning_string + rule.return_warnings()

        return warning_string

    def return_rule_warnings_for_cmlt(self):
        """
        Loops throught the self.yara_rules array and returns a string for of warnings in cmlt format
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_rule_warnings_for_cmlt() is deprecated, '
            'use YaraFileProcessor.return_file_warnings_for_cmlt() instead',
            DeprecationWarning
        )
        warning_string = ''

        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_string = warning_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                    warning_string = warning_string + rule.return_warnings_for_cmlt()

        return warning_string

    def return_file_warnings_for_cmlt(self):
        """
        Loops throught the self.yara_rules array and returns a string for of warnings in cmlt format
        :return:
        """
        warning_string = ''

        for rule in self.yara_rules:
            if rule.rule_return:
                if rule.return_warning():
                    warning_string = warning_string + "{indent:>8} {name:10}".format(indent="- ",
                                                                                     name=rule.get_rule_name() + ":\n")
                    warning_string = warning_string + rule.return_warnings_for_cmlt()

        return warning_string

    def return_original_rule(self):
        """
        Returns the original rule string
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_original_rule() is deprecated, '
            'use YaraFileProcessor.return_original_file() instead',
            DeprecationWarning
        )
        return self.original_rule_string

    def return_original_file(self):
        """
        Returns the original rule string
        :return:
        """
        return self.original_rule_string

    def return_edited_rule(self):
        """
        Returns the edited rule string
        :return:
        """
        import warnings
        warnings.warn(
            'YaraFileProcessor.return_original_rule() is deprecated, '
            'use YaraFileProcessor.return_original_file() instead',
            DeprecationWarning
        )
        return self.edited_rule_string

    def return_edited_file(self):
        """
        Returns the edited rule string
        :return:
        """
        return self.edited_rule_string



class YaraRule:
    """
    YaraRule objects contain a string representation of a rule, a plyara representation of the rule and the RuleReturn
        object
    """

    def __init__(self, rule_string, rule_plyara):
        self.rule_string = rule_string
        self.rule_plyara = rule_plyara
        self.rule_name = str(self.rule_plyara.get('rule_name'))
        self.rule_return = YaraReturn(rule_string)

    def add_rule_return(self, rule_return):
        self.rule_return = rule_return

    def return_error(self):
        return self.rule_return.error_state()

    def return_errors(self):
        error_string = ''

        if isinstance(self.rule_return, YaraReturn):
            if self.rule_return.error_state():
                error_string = self.rule_return.return_errors()
                if error_string:
                    error_string = error_string + '\n'
        else:
            if not self.rule_return.rule_validity:
                error_string = self.rule_return.return_errors()
                if error_string:
                    error_string = error_string + '\n'

        return error_string

    def return_errors_for_cmlt(self):
        error_string = ''

        if isinstance(self.rule_return, YaraReturn):
            if self.rule_return.error_state():
                error_string = self.rule_return.return_errors_for_cmlt()
                if error_string:
                    error_string = error_string + '\n'
        else:
            if not self.rule_return.rule_validity:
                error_string = self.rule_return.return_errors_for_cmlt()
                if error_string:
                    error_string = error_string + '\n'

        return error_string

    def return_warning(self):
        return self.rule_return.warning_state()

    def return_warnings(self):
        warning_string = ''
        if self.rule_return.warning_state():
            warning_string = self.rule_return.return_warnings()
            if warning_string:
                warning_string = warning_string + '\n'

        return warning_string

    def return_warnings_for_cmlt(self):
        warning_string = ''
        if self.rule_return.warning_state():
            warning_string = self.rule_return.return_warnings_for_cmlt()
            if warning_string:
                warning_string = warning_string + '\n'

        return warning_string

    def return_rule_return(self):
        return self.rule_return

    def get_rule_name(self):
        return self.rule_name

    def return_original_rule(self):
        return self.rule_return.return_original_rule()

    def return_edited_rule(self):
        return self.rule_return.edited_rule()


class YaraReturn:
    """
    YaraReturn class used to pass the error state of the processed rules, what metadata tags have issues,
        a string representation of the original rule and if the rule has no errors a string representation of the rule
        with all the created or changed metadata tags, etc.
    """
    def __init__(self, original_rule):
        # Overall rule error flag
        self.rule_errors = False
        # Overall warning flag
        self.rule_warnings = False
        # collection of all the errors
        self.errors = collections.OrderedDict()
        # collection of all the warnings
        self.warnings = collections.OrderedDict()
        # the original_rule
        self.original_rule = original_rule
        # set
        self.edited_rule = None

    def update_error(self, rule_error, error_tag, message):
        if not self.rule_errors:
           self.rule_errors = rule_error

        self.errors[error_tag] = message

    def update_warning(self, rule_warning, warning_tag, message):
        if not self.rule_warnings:
            self.rule_warnings = rule_warning

        self.warnings[warning_tag] = message

    def __build_return_string(self, collection):
        return_string = ''
        for index, metadata in enumerate(collection):
            if index > 0:
                return_string += '\n'
            return_string = '{}{}: {}'.format(return_string, metadata, collection[metadata])

        return return_string

    def __build_return_string_cmlt(self, collection):
        format_string = '{indent:>9} {key:30} {value}'
        return_string = '\n'.join([
            format_string.format(indent='-', key=k + ':', value=v)
            for k, v in collection.items()
        ])

        return return_string

    def error_state(self):
        return self.rule_errors

    def return_errors(self):
        error_string = ''
        if self.rule_errors:
            error_string = self.__build_return_string(self.errors)

        return error_string

    def return_errors_for_cmlt(self):
        error_string = ''
        if self.rule_errors:
            error_string = self.__build_return_string_cmlt(self.errors)

        return error_string

    def warning_state(self):
        return self.rule_warnings

    def return_warnings(self):
        warning_string = ''
        if self.rule_warnings:
            warning_string = self.__build_return_string(self.warnings)

        return warning_string

    def return_warnings_for_cmlt(self):
        warning_string = ''
        if self.rule_warnings:
            warning_string = self.__build_return_string_cmlt(self.warnings)

        return warning_string

    def return_original_rule(self):
        return self.original_rule

    def return_edited_rule(self):
        return self.edited_rule

    def set_edited_rule(self, edited_rule):
        self.edited_rule = edited_rule

    def return_validated_rule(self):
        """
        Created to duplicate functionality from original YaraValidatorReturn class
        :return: returns the self.edited_rule instead of the original self.validated_rule
        """
        return self.edited_rule

    def set_validated_rule(self, valid_rule):
        """
        Created to duplicate functionality from original YaraValidatorReturn class
        :param valid_rule:
        :return:
        """
        self.edited_rule = valid_rule

    def __find_meta_start_end(self, rule_to_process):
        """
        A string representation of a YARA rule is passed into this function, it performs the splitlines() function,
            searches for the start and the end indexes of the meta section of the first YARA rule.
        :param rule_to_process: The Rule to be processed
        :return: a tuple of the array of lines for the rule processed, and a list of start and end of meta indices
        """
        rule_to_process_lines = rule_to_process.splitlines()
        meta_offsets = []
        meta_start = 0
        meta_end = 0
        meta_regex = r'^\s*meta\s*:\s*$'
        next_section = r'^\s*(?:strings|condition)\s*:\s*$'

        for index, line in enumerate(rule_to_process_lines):
            if re.match(meta_regex, line):
                meta_start = index
            elif re.match(next_section, line) and meta_start > 0:
                meta_end = index
                break

        return rule_to_process_lines, meta_start, meta_end

    def rebuild_rule(self):
        """
        Rebuilds the rule if there are no errors and as long as there are changes. This was created to maintain
            any comments outside of the metadata section
        :return: No return
        """
        if self.edited_rule[-1] == '\n':
            self.edited_rule = self.edited_rule[:-1]

        if self.original_rule is None or self.edited_rule is None:
            exit()
        elif self.original_rule == self.edited_rule:
            return

        yara_original_lines, yara_original_meta_start, yara_original_meta_end = self.__find_meta_start_end(self.original_rule)
        yara_edited_lines, yara_edited_meta_start, yara_edited_meta_end = self.__find_meta_start_end(self.edited_rule)

        yara_new_file = []
        if yara_original_meta_start != 0 and yara_original_meta_end != 0 and yara_edited_meta_start != 0 and yara_edited_meta_end != 0:
            yara_new_file = yara_original_lines[0:yara_original_meta_start] + \
                            yara_edited_lines[yara_edited_meta_start:yara_edited_meta_end] + \
                            yara_original_lines[yara_original_meta_end:]
            yara_new_file = '\n'.join(yara_new_file)

        if self.original_rule != yara_new_file:
            self.edited_rule = yara_new_file



import os
from pathlib import Path

# set current working directory
SCRIPT_LOCATION = Path(__file__).resolve().parent
VALIDATOR_CFG = os.environ.get('VALIDATOR_CFG', SCRIPT_LOCATION / 'validator_cfg.yml')

# Allow use of custom paths/configurations besides the defaults
CONFIG_YAML_PATH = os.environ.get('CONFIG_YAML_PATH', SCRIPT_LOCATION / 'CCCS_YARA.yml')
CONFIG_VALUES_YAML_PATH = os.environ.get('CONFIG_VALUES_YAML_PATH', SCRIPT_LOCATION / 'CCCS_YARA_values.yml')
MITRE_STIX_DATA_PATH = os.environ.get('MITRE_STIX_DATA_PATH', SCRIPT_LOCATION / 'cti')



import stix2.utils
from stix2 import Filter
from datetime import datetime

# add the casefold to the supported filter operations
from stix2.datastore.filters import FILTER_OPS
FILTER_OPS.append('casefold')


class FilterCasefold(Filter):
    def _check_property(self, stix_obj_property):
        """Check a property of a STIX Object against this casefold filter, but with a casefold operator.
        Args:
            stix_obj_property: value to check this filter against
        Returns:
            True if property matches the filter,
            False otherwise.
        """
        # Had to keep the following code so that
        # If filtering on a timestamp property and the filter value is a string,
        # try to convert the filter value to a datetime instance.
        if isinstance(stix_obj_property, datetime) and isinstance(self.value, str):
            filter_value = stix2.utils.parse_into_datetime(self.value)
        else:
            filter_value = self.value

        if self.op == 'casefold':
            return stix_obj_property.casefold() == filter_value.casefold()
        else:
            Filter._check_property(self, stix_obj_property)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fh:
    requirements = fh.readlines()

s = setup(
    name="cccs-yara",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="A CCCS utility for YARA rule metadata validation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=requirements,
    entry_points="""
    [console_scripts]
    yara_validator=yara_validator.cli:main
    """,
    package_data={'yara_validator': ['*.yml']},
    include_package_data=True,
)
